local setmetatable = setmetatable
local print = print

local mmf = mmf

module (...)

-- ----------------------------------------------------------------------------
-- Generic Control

Control = {
	exportID = 0,
	fixedID = 0,
	x = 0, _x = 0,
	y = 0, _y = 0,
	w = 0, _w = 0,
	h = 0, _h = 0,
	clientx = nil,
	clienty = nil,
	clientw = nil,
	clienth = nil,
	enabled = true,
	checked = false,
	enagable = false,
	engaged = false,
	engageX = 0,
	engageY = 0,
}

function ControlObject (obj)
	return Control:new(obj)
end

function Control:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Control:Bind (obj)
	self.exportID = obj
	self.fixedID = mmf.Object.GetFixed (obj)
end

function Control:SetPosition (x, y)
	if (x ~= nil) then
		self.x = x
	end
	if (y ~= nil) then
		self.y = y
	end
	
	if (self.x == self._x and self.y == self._y) then
		return
	end
	
	mmf.Object.SetPosition(self.exportID, self.x, self.y)
	self._x, self._y = self.x, self.y
end

function Control:DeferSetPosition (x, y)
	if (x ~= nil) then
		self.x = x
	end
	if (y ~= nil) then
		self.y = y
	end
		
	if (self.x == self._x and self.y == self._y) then
		--return
	end
	
	mmf.Object.SetPosition(self.exportID, self._x, self._y)
	self._x, self._y = self.x, self.y
end

function Control:SetSize (w, h)
	if (w ~= nil) then
		self.w = w
	end
	if (h ~= nil) then
		self.h = h
	end
	
	if (self.w == self._w and self.h == self._h) then
		return
	end
	
	mmf.Control.SetSize(self.fixedID, self.w, self.h)
	self._w, self._h = self.w, self.h
end

function Control:SetLayer (layer)
	mmf.Object.SetLayer(self.exportID, layer)
end

function Control:SetText (text)
	mmf.Control.SetText(self.fixedID, text)
end

function Control:SetColor (color)
	mmf.Control.SetColor(self.fixedID, color)
end

function Control:UnderMouse (curs)
	if (curs.x >= self.x and curs.x < self.x + self.w and curs.y >= self.y and curs.y < self.y + self.h) then
		return true
	end
	return false
end

function Control:ClientUnderMouse (curs)
	if (self.clientx == nil or self.clienty == nil or self.clientw == nil or self.clienth == nil) then
		return self:UnderMouse(curs)
	end
	
	if (curs.x >= self.clientx and curs.x < self.clientx + self.clientw and 
		curs.y >= self.clienty and curs.y < self.clienty + self.clienth) then
		return true
	end
	return false
end

-- TestEngage
-- ------------------------------------
-- Tests when a user explicitly clicks and holds the mouse button down over an object; facilitates dragging.
-- Sets engaged = true when the object is "held" by the cursor.
-- Sets engagable = true when the cursor is over the object and not "holding" anything else.

function Control:TestEngage (curs)
	-- Can we currently engage the control?
	local overDiv = self:ClientUnderMouse(curs)
	if (curs.LDown == false and overDiv) then
		self.engagable = true
	end
	
	-- On click, engage the control
	if (self.engagable and curs.LDown) then
		self.engaged = true
		self.engageX = curs.x
		self.engageY = curs.y
	elseif (self.engaged and curs.LDown == false) then
		self.engaged = false
	end
	
	-- Cancel engagable status if necessary
	if (curs.LDown or overDiv == false) then
		self.engagable = false
	end
end

-- Enable
-- ------------------------------------
-- For controls that support it, puts a control in an enabled state if it isn't already

function Control:Enable ()
	if (self.enabled == false) then
		self.enabled = true
		mmf.Control.Enable(self.fixedID)
	end
end

-- Disable
-- ------------------------------------
-- For controls that support it, puts a control in a disabled state if it isn't already

function Control:Disable ()
	if (self.enabled) then
		self.enabled = false
		mmf.Control.Disable(self.fixedID)
	end
end

-- Check
-- ------------------------------------
-- For controls that support it, puts a control in a checked state if it isn't already

function Control:Check ()
	if (self.checked == false) then
		self.checked = true
		mmf.Control.Check(self.fixedID)
	end
end

-- Uncheck
-- ------------------------------------
-- For controls that support it, removes a control's checked state if it isn't already

function Control:Uncheck ()
	if (self.checked) then
		self.checked = false
		mmf.Control.Uncheck(self.fixedID)
	end
end

function Control:Show ()
	--if (mmf.Object.GetVisibility(self.exportID) == false) then
		mmf.Object.SetVisibility(self.exportID, true)
	--end
end

function Control:Hide ()
	--if (mmf.Object.GetVisibility(self.exportID) == true) then
		mmf.Object.SetVisibility(self.exportID, false)
	--end
end

function Control:Visible ()
	return mmf.Object.GetVisibility(self.exportID)
end

-- OnButtonPress
-- ------------------------------------
-- Overridable function to run any tasks when the control is pressed (if the control is a button)

function Control:OnButtonPress () end

-- Init
-- ------------------------------------
-- Overridable function to run any tasks on this control that should be performed during
-- initialization only.

function Control:Init () end

-- Update
-- ------------------------------------
-- Overridable function to perform any updates to this control during the dispay cycle.
-- By default, a control's size and position are kept up to date on the frame

function Control:Update ()
	self:SetPosition()
	self:SetSize()
end

-- Step
-- ------------------------------------
-- Overridable function to run tasks each event loop.  Step is run before Update.

function Control:Step () end

-- ----------------------------------------------------------------------------
-- Scrollbar Control

Scrollbar = Control:new( {
	position = 0,
	rangeMin = 0,
	rangeMax = 0,
} )

function ScrollbarObject (obj)
	return Scrollbar:new(obj)
end

function Scrollbar:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Scrollbar:SetScrollPosition (pos)
	if (pos ~= nil) then
		if (pos == self.position) then
			return
		end
		self.position = pos
	end
	
	mmf.Scrollbar.SetScrollPosition(self.fixedID, self.position)
end

function Scrollbar:SetScrollRange (min, max)
	if (min ~= nil) then
		self.rangeMin = min
	end
	if (max ~= nil) then
		self.rangeMax = max
	end
	
	mmf.Scrollbar.SetScrollRange(self.fixedID, self.rangeMin, self.rangeMax)
end

-- ----------------------------------------------------------------------------
-- Active Picture (APO) Control

ActivePicture = Control:new( {
	image = '',
	wSource = 0,
	hSource = 0,
} )

function ActivePictureObject (obj)
	return ActivePicture:new(obj)
end

function ActivePicture:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function ActivePicture:LoadImage (img)
	if (img ~= nil) then
		self.image = img
	end
	
	self.wSource, self.hSource = mmf.APO.LoadImage(self.fixedID, self.image)
end

function ActivePicture:PasteBG (dx, dy, sx, sy, sw, sh)
	mmf.APO.CreateBG(self.fixedID, dx, dy, sx, sy, sw, sh)
end

-- ----------------------------------------------------------------------------
-- Clipped Image Control

Clipped = Control:new( {
	image = '', _image = '',
	xOff = 0, _xOff = 0,
	yOff = 0, _yOff = 0,
	wSource = 0,
	hSource = 0,
} )

function ClippedObject (obj)
	return Clipped:new(obj)
end

function Clipped:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Clipped:LoadImage (img)
	if (img ~= nil) then
		self.image = img
	end
	
	if (self.image == self._image) then
		return
	end

	self.wSource, self.hSource = mmf.Clipped.LoadImage(self.fixedID, self.image)
	self._image = self.image
end

function Clipped:SetOffset (x, y)
	if (x ~= nil) then
		self.xOff = x
	end
	if (y ~= nil) then
		self.yOff = y
	end
	
	if (self.xOff == self._xOff and self.yOff == self._yOff) then
		return
	end
	
	mmf.Clipped.SetOffset(self.fixedID, self.xOff, self.yOff)
	self._xOff, self._yOff = self.xOff, self.yOff
end

-- ----------------------------------------------------------------------------
-- Grid Control

Grid = Control:new( {
	wCell = 0, _wCell = 0,
	hCell = 0, _hCell = 0,
	xOff = 0,
	yOff = 0,
	mode = 0,
	color = 0,
	visible = true,
} )

function GridObject (obj)
	return Grid:new(obj)
end

function Grid:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Grid:SetCellSize (cellw, cellh)
	if (cellw ~= nil) then
		self.wCell = cellw
	end
	if (cellh ~= nil) then
		self.hCell = cellh
	end
	
	if (self.wCell == self._wCell and self.hCell == self._hCell) then
		return
	end
	
	mmf.Grid.SetCellSize(self.fixedID, self.wCell, self.hCell)
	self._wCell, self._hCell = self.wCell, self.hCell
end

function Grid:SetOffset (offx, offy)
	if (offx ~= nil) then
		self.xOff = offx
	end
	if (offy ~= nil) then
		self.yOff = offy
	end
	
	mmf.Grid.SetOffset(self.fixedID, self.xOff, self.yOff)
end

function Grid:SetColor (color)
	if (color ~= nil) then
		self.color = color
	end
	
	mmf.Grid.SetLineColor(self.fixedID, self.color)
end

function Grid:SetMode (mode)
	local oldm = self.mode
	if (mode ~= nil) then
		self.mode = mode
	end
	
	if (self.mode == oldm) then
		return
	end
	
	if (self.mode == 0) then
		mmf.Grid.SetDotMode(self.fixedID)
	elseif (self.mode == 1) then
		mmf.Grid.SetLineMode(self.fixedID)
	end
end

function Grid:Show ()
	if (self.visible) then
		return
	end
	
	mmf.Control.Show(self.fixedID)
	self.visible = true
end

function Grid:Hide ()
	if (not self.visible) then
		return
	end
	
	mmf.Control.Hide(self.fixedID)
	self.visible = false
end

-- ----------------------------------------------------------------------------
-- Viewport

Viewport = Control:new( {
	xSource = 0, _xSource = 0,
	ySource = 0, _ySource = 0,
	wSource = 0, _wSource = 0,
	hSource = 0, _hSource = 0,
	zoomLevel = 0,
} )

function ViewportObject (obj)
	return Viewport:new(obj)
end

function Viewport:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Viewport:SetSource (x, y, w, h)
	if (x ~= nil) then
		self.xSource = x
	end
	if (y ~= nil) then
		self.ySource = y
	end
	if (w ~= nil) then
		self.wSource = w
	end
	if (h ~= nil) then
		self.hSource = h
	end

	if (self.xSource == self._xSource and self.ySource == self._ySource and 
		self.wSource == self._wSource and self.hSource == self._hSource) then
		return
	end
	
	mmf.Viewport.SetSource(self.fixedID, self.xSource, self.ySource, self.wSource, self.hSource)
	self._xSource, self._ySource, self._wSource, self._hSource = self.xSource, self.ySource, self.wSource, self.hSource
end

-- ----------------------------------------------------------------------------
-- Cursor

Cursor = {
	exportID = 0,
	fixedID = 0,
	x = 0,
	y = 0,
	state = "win",
	stateTable = { },
}

function CursorObject (obj)
	return Cursor:new(obj)
end

function Cursor:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Cursor:Bind (obj)
	self.exportID = obj
	self.fixedID = mmf.Object.GetFixed (obj)
end

function Cursor:SetPosition (x, y)
	if (x ~= nil) then
		self.x = x
	end
	if (y ~= nil) then
		self.y = y
	end
	
	mmf.Object.SetPosition(self.exportID, self.x, self.y)
end

function Cursor:SetState (state)
	self.state = state
	
	if (state == "win") then
		mmf.Mouse.Show()
		mmf.Object.SetVisibility(self.exportID, false)
	else
		mmf.Mouse.Hide()
		mmf.Object.SetVisibility(self.exportID, true)
		mmf.Object.SetAnimation(self.exportID, self.stateTable[state])
		mmf.Object.Update(self.exportID)
	end
end

function Cursor:Update ()
	self:SetPosition()
end