local CLASS = require "./lua/editor_com"
local Container = require "./lua/container"
local Undo = require "./lua/undo"

local util = require "./lua/util"

local Base = require "./lua/editor_baselayer"
local Tile = require "./lua/tile"

local Profile = Profile

module(..., package.seeall)

-- ----------------------------------------------------------------------------
-- Unbound objects from the editor

Controls = { }
Current = nil
Cursor = { }
FixedTable = { }

-- ----------------------------------------------------------------------------

Exports = {
	tilePaletteSource = 0,
	tilePalette = 0,
	tileSelect = 0,
}

Layer = Base.Layer:new( {
	map = { },
	float = { },
	stamp = false,
	
	last_tx = -1,
	last_ty = -1,
	undoStack = nil,
	undoPool = nil,
	latch = false,
	
	tileset = '',
	typeid = 0,
	tileWidth = 0,
	tileHeight = 0,
	xSpacing = 0,
	ySpacing = 0,
	xOrigin = 0,
	yOrigin = 0,
	xSelectStart = 0,
	ySelectStart = 0,
	xSelectEnd = 0,
	ySelectEnd = 0,
} )

function Layer:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	obj:Construct()
	return obj
end

function Layer:Construct ()
	self.layerType = 1
	self.float = { }
	self.undoStack = Undo.UndoStack:New()
	self.undoPool = nil
	
	self.curTool = Profile.startTools.tile
	
	if (self.subtype == 'normal') then
		self.typeid = 1
	elseif (self.subtype == 'collision') then
		self.typeid = 2
	elseif (self.subtype == 'platform') then
		self.typeid = 3
	end
		
	-- Setup defaults - TODO: Move to platform profile
	
	self.tileWidth = 16
	self.tileHeight = 16
	self.xSpacing = 1
	self.ySpacing = 1
	self.xOrigin = 0
	self.yOrigin = 0
	--self.tileset = './data/chip1.png'
	
	self.map = Tile.TileMap:new( {
		layer = self, 
		xMax = math.ceil(Frame.frameWidth / self.tileWidth), 
		yMax = math.ceil(Frame.frameHeight / self.tileHeight)
	} )

	--Construct controls
	
	self.controls.tilePaletteSource = CLASS.ActivePictureObject( {parent = self} )
	self.controls.tilePalette = CLASS.ClippedObject( {parent = self} )
	self.controls.tileSelect = CLASS.ControlObject( {parent = self} )
	
	self.tools.tbSelect = CLASS.ControlObject( {parent = self, w = 28, h = 28, name = 'select'} )
	self.tools.tbDraw = CLASS.ControlObject( {parent = self, w = 28, h = 28, name = 'draw'} )
	self.tools.tbErase = CLASS.ControlObject( {parent = self, w = 28, h = 28, name = 'erase'} )
	self.tools.tbStamp = CLASS.ControlObject( {parent = self, w = 28, h = 28, name = 'stamp'} )
	self.tools.tbFill = CLASS.ControlObject( {parent = self, w = 28, h = 28, name = 'fill'} )
	
	-- Override display functions for Tile-editor controls
	
	function self.controls.tilePalette:Update ()
		self.x = Controls.sidebar.clientx
		self.y = Controls.sidebar.clienty
		self.w = Controls.sidebar.clientw
		self.h = Controls.sidebar.clienth
		
		self.xOff = Controls.sidebarHScroll.position + self.parent.xOrigin
		self.yOff = Controls.sidebarVScroll.position + self.parent.yOrigin
		
		local xoverflow = self.wSource - self.w
		local yoverflow = self.hSource - self.h
		
		-- Update scrollbars
		if (xoverflow <= 0) then
			Controls.sidebarHScroll:Disable()
		else
			Controls.sidebarHScroll:Enable()
			Controls.sidebarHScroll:SetScrollRange(0, xoverflow)
		end
		
		if (yoverflow <= 0) then
			Controls.sidebarVScroll:Disable()
		else
			Controls.sidebarVScroll:Enable()
			Controls.sidebarVScroll:SetScrollRange(0, yoverflow)
		end
		
		self:SetPosition()
		self:SetSize()
		self:SetOffset()
	end
	
	function self.controls.tileSelect:Update ()
		local selectWidth = self.parent.xSelectEnd - self.parent.xSelectStart + 1
		local selectHeight = self.parent.ySelectEnd - self.parent.ySelectStart + 1
		
		self.x = Controls.sidebar.clientx + self.parent.xSelectStart * (self.parent.tileWidth + self.parent.xSpacing) - (Controls.sidebarHScroll.position - self.parent.xOrigin)
		self.y = Controls.sidebar.clienty + self.parent.ySelectStart * (self.parent.tileHeight + self.parent.ySpacing) - (Controls.sidebarVScroll.position - self.parent.yOrigin)
		self.w = selectWidth * (self.parent.tileWidth + self.parent.xSpacing)
		self.h = selectHeight * (self.parent.tileHeight + self.parent.ySpacing)
		
		-- Don't draw the control out of range
		local x = math.min(self.x, Controls.sidebarVScroll.x + 8)
		local y = math.min(self.y, Controls.sidebarHScroll.y + 8)
		local w = math.min(self.w, math.max(0, Controls.sidebarVScroll.x + 8 - self.x))
		local h = math.min(self.h, math.max(0, Controls.sidebarHScroll.y + 8 - self.y))
		
		self:DeferSetPosition(x, y)
		self:SetSize(w, h)
	end
	
	-- Override step funtions for Tile-editor controls
	
	function self.controls.tileSelect:Step ()
		local tpal = self.parent.controls.tilePalette
		
		if (win.state.menusOpen > 0 or win.state.active == 0) then
			return
		end
		
		tpal:TestEngage(Cursor)
		
		if (tpal.engaged) then
			local xOrigin = tpal.x - self.parent.xOrigin - Controls.sidebarHScroll.position
			local yOrigin = tpal.y - self.parent.yOrigin - Controls.sidebarVScroll.position
			
			local xOffStart = tpal.engageX - xOrigin
			local yOffStart = tpal.engageY - yOrigin
			local xOffEnd = Cursor.x - xOrigin
			local yOffEnd = Cursor.y - yOrigin
			
			-- Keep start in front of end
			if (xOffStart >= xOffEnd) then
				xOffStart, xOffEnd = xOffEnd, xOffStart
			end
			if (yOffStart >= yOffEnd) then
				yOffStart, yOffEnd = yOffEnd, yOffStart
			end
			
			-- Keep stuff in bounds
			if (xOffStart < 0) then
				xOffStart = 0
			end
			if (yOffStart < 0) then
				yOffStart = 0
			end
			
			if (xOffEnd >= tpal.wSource - 1) then
				xOffEnd = tpal.wSource - 1
			end
			if (yOffEnd >= tpal.hSource - 1) then
				yOffEnd = tpal.hSource - 1
			end
			
			-- Convert to tile coordinates
			self.parent.xSelectStart = math.floor(xOffStart / (self.parent.tileWidth + self.parent.xSpacing))
			self.parent.ySelectStart = math.floor(yOffStart / (self.parent.tileHeight + self.parent.ySpacing))
			self.parent.xSelectEnd = math.floor(xOffEnd / (self.parent.tileWidth + self.parent.xSpacing))
			self.parent.ySelectEnd = math.floor(yOffEnd / (self.parent.tileHeight + self.parent.ySpacing))
			
			self:Update()
		end
	end

	-- Toolbar callback
	
	function self.tools.tbSelect:OnButtonPress()
		self.parent.curTool = 'select'
		self.parent.cursor = 'win'
		self:Check()
		self.parent.tools.tbDraw:Uncheck()
		self.parent.tools.tbErase:Uncheck()
		self.parent.tools.tbStamp:Uncheck()
		self.parent.tools.tbFill:Uncheck()
	end
	
	function self.tools.tbDraw:OnButtonPress()
		self.parent.curTool = 'draw'
		self.parent.cursor = 'draw'
		self:Check()
		self.parent.tools.tbSelect:Uncheck()
		self.parent.tools.tbErase:Uncheck()
		self.parent.tools.tbStamp:Uncheck()
		self.parent.tools.tbFill:Uncheck()
	end
	
	function self.tools.tbErase:OnButtonPress()
		self.parent.curTool = 'erase'
		self.parent.cursor = 'erase'
		self:Check()
		self.parent.tools.tbSelect:Uncheck()
		self.parent.tools.tbDraw:Uncheck()
		self.parent.tools.tbStamp:Uncheck()
		self.parent.tools.tbFill:Uncheck()
	end
	
	function self.tools.tbStamp:OnButtonPress()
		self.parent.curTool = 'stamp'
		self.parent.cursor = 'win'
		self:Check()
		self.parent.tools.tbSelect:Uncheck()
		self.parent.tools.tbDraw:Uncheck()
		self.parent.tools.tbErase:Uncheck()
		self.parent.tools.tbFill:Uncheck()
	end
	
	function self.tools.tbFill:OnButtonPress()
		self.parent.curTool = 'fill'
		self.parent.cursor = 'fill'
		self:Check()
		self.parent.tools.tbSelect:Uncheck()
		self.parent.tools.tbDraw:Uncheck()
		self.parent.tools.tbErase:Uncheck()
		self.parent.tools.tbStamp:Uncheck()
	end
end

function Layer:Init ()
	self.controls.tilePaletteSource:Bind(Exports.tilePaletteSource)
	self.controls.tilePalette:Bind(Exports.tilePalette)
	self.controls.tileSelect:Bind(Exports.tileSelect)
	
	self.tools.tbSelect:Bind(Base.Exports.tbSelect)
	self.tools.tbDraw:Bind(Base.Exports.tbDraw)
	self.tools.tbErase:Bind(Base.Exports.tbErase)
	self.tools.tbStamp:Bind(Base.Exports.tbStamp)
	self.tools.tbFill:Bind(Base.Exports.tbFill)
end

function Layer:Activate ()
	if (self.tileset ~= '') then
		self.controls.tilePalette:Show()
		self.controls.tileSelect:Show()
		
		for k,v in pairs(self.tools) do
			v:Disable()
			v:Enable()
		end
	else
		self.curTool = ''
	end

	self.controls.tilePalette:LoadImage(Frame.dataDir .. self.tileset)
	self.controls.tilePaletteSource:LoadImage(Frame.dataDir .. self.tileset)
	self.controls.tilePaletteSource:SetSize(self.controls.tilePaletteSource.wSource, self.controls.tilePaletteSource.hSource)
	self.controls.tilePaletteSource:SetLayer(Conf.baseLayer + self.order)
	
	-- Sidebar adjustment
	local width = math.min(self.controls.tilePalette.wSource, math.floor(mmf.Window.GetCWidth() / 4))
	Controls.sidebar.w = math.max(Controls.sidebarHScroll.h * 2, width) + Controls.sidebarVScroll.w + 2
	
	-- Rebind controls
	FixedTable[self.tools.tbSelect.fixedID] = self.tools.tbSelect
	FixedTable[self.tools.tbDraw.fixedID] = self.tools.tbDraw
	FixedTable[self.tools.tbErase.fixedID] = self.tools.tbErase
	FixedTable[self.tools.tbStamp.fixedID] = self.tools.tbStamp
	FixedTable[self.tools.tbFill.fixedID] = self.tools.tbFill

	-- Why is this necessary?
	self.controls.tileSelect.parent = self
	self.tools.tbSelect.parent = self
	self.tools.tbDraw.parent = self
	self.tools.tbErase.parent = self
	self.tools.tbStamp.parent = self
	self.tools.tbFill.parent = self
	
	for k,v in pairs(self.tools) do
		if (self.curTool == v.name and self.tileset ~= '') then
			v:OnButtonPress()
		end
	end
end

function Layer:Deactivate ()
	self.controls.tilePalette:Hide()
	self.controls.tileSelect:Hide()
	
	self.controls.tilePalette:LoadImage('')
	self.controls.tilePaletteSource:LoadImage('')
	self.controls.tilePalette.wSource = self.controls.tilePalette.w
	self.controls.tilePalette.hSource = self.controls.tilePalette.h
	
	for k,v in pairs(self.tools) do
		v:SetPosition(0, -64)
		v:Disable()
		v:Uncheck()
	end
	
	Controls.sidebarHScroll:Disable()
	Controls.sidebarVScroll:Disable()
end

function Layer:Reload ()
	for k,v in pairs(self.tools) do
		v:Disable()
	end
	self:Activate()
	
	self.map:PaintAt(0, 0)
end

function Layer:Reset ()
	self:Deactivate()
	
	self.map:ClearMap()
end

function Layer:TileProperties ()
	return {
		tileWidth = self.tileWidth,
		tileHeight = self.tileHeight,
	}
end

function Layer:OpenProperties ()
	self.tempdata = { }
	mmf.Dialog.OpenTileProperties()
end

function Layer:UpdateProperties ()
	self.map:ClearMap()
	
	self.xSpacing = self.tempdata.xSpacing
	self.ySpacing = self.tempdata.ySpacing
	self.xOrigin = self.tempdata.xOrigin
	self.yOrigin = self.tempdata.yOrigin
	--self.tileset = "./data/jungle.jpg"
	self.tileWidth = self.tempdata.tileWidth
	self.tileHeight = self.tempdata.tileHeight
	
	local dst = util.basename(self.tempdata.tileset)
	if (self.tileset ~= dst) then
		util.filecopy(self.tempdata.tileset, Frame.dataDir .. dst)
		self.tileset = dst
	end
	
	self:Activate()

	self.map:PaintAt(0, 0)
	self.queuedUpdate = false
	
	editor.SetModified()
end

function Layer:Redraw ()
	self.map:ClearArea(0, 0, self.map.xMax, self.map.yMax)
	self.map:PaintAt(0, 0)
end

function EnableCutCopy ()
	Controls.toolbar.tbCut:Enable()
	mmf.Menu.Enable("EditCut")
	Controls.toolbar.tbCopy:Enable()
	mmf.Menu.Enable("EditCopy")
end

function DisableCutCopy ()
	Controls.toolbar.tbCut:Disable()
	mmf.Menu.Disable("EditCut")
	Controls.toolbar.tbCopy:Disable()
	mmf.Menu.Disable("EditCopy")
end

function EnablePaste ()
	Controls.toolbar.tbPaste:Enable()
	mmf.Menu.Enable("EditPaste")
end

function DisablePaste ()
	Controls.toolbar.tbPaste:Disable()
	mmf.Menu.Disable("EditPaste")
end

function Layer:Cut ()
	if (self.curTool ~= 'select' or Controls.floatingBox.active == false) then
		return
	end
	
	self:ClearFloat()
	
	self.clipboard = self.float
	self.float = nil
	
	Controls.floatingBox:Release()
	Controls.floatingBox:Stop()
	
	DisableCutCopy()
	
	self.undoStack:LogData(self.undoPool)
	self.undoPool = nil
end

function Layer:Copy ()
	if (self.curTool ~= 'select' or Controls.floatingBox.active == false) then
		return
	end
	
	self.clipboard = self.float
end

function Layer:Paste ()
	if (self.curTool ~= 'select' or self.clipboard == nil) then
		return
	end
	
	if (Controls.floatingBox.active and self.float) then
		self:FloatUndo(1)
		self:PaintFloat()
		self.map:MergeMap(self.float, Controls.floatingBox.sx, Controls.floatingBox.sy)
	end
	
	self.float = self.clipboard
	
	Controls.floatingBox:Stop()
	Controls.floatingBox:StartSnapping(Cursor.canvas.xtile + 5, Cursor.canvas.ytile + 5, self.float.xMax + 1, self.float.yMax + 1)
	self:PaintFloat()
end

function Layer:UpdateControls ()
	self.controls.tilePalette:Update()
	self.controls.tileSelect:Update()
end

function Layer:UpdateTools (x, y)
	-- Button List
	local buttons = {
		'tbSelect', 'tbDraw', 'tbErase', 'tbStamp', 'tbFill',
	}
	
	-- Update button positions
	self.tools.tbSelect.x = x
	self.tools.tbDraw.x = self.tools.tbSelect.x + self.tools.tbSelect.w
	self.tools.tbErase.x = self.tools.tbDraw.x + self.tools.tbDraw.w
	self.tools.tbStamp.x = self.tools.tbErase.x + self.tools.tbErase.w
	self.tools.tbFill.x = self.tools.tbStamp.x + self.tools.tbStamp.w
	
	for k,v in pairs(buttons) do
		self.tools[v].y = y
		self.tools[v]:SetPosition()
	end
end

function Layer:UpdateDimensions ()
	local nxMax = math.ceil(Frame.frameWidth / self.tileWidth)
	local nyMax = math.ceil(Frame.frameHeight / self.tileHeight)
	
	if (nxMax < self.map.xMax) then
		self.map:RemoveArea(nxMax, 0, self.map.xMax, self.map.yMax)
	end
	if (nyMax < self.map.yMax) then
		self.map:RemoveArea(0, nyMax, self.map.xMax, self.map.yMax)
	end
	
	self.map.xMax = nxMax
	self.map.yMax = nyMax
end

function Layer:Step ()
	if (win.state.menusOpen > 0 or win.state.active == 0) then
		return
	end
	
	if (self.queuedUpdate) then
		self:UpdateProperties()
	end
	
	self:UpdateDimensions()

	Controls.clientArea:TestEngage(Cursor)
	
	-- Update cursor
	if (Controls.clientArea.engagable or Controls.clientArea.engaged) then
		Cursor:SetState(self.cursor)
	else
		Cursor:SetState('win')
	end
	
	-- Undo state
	if (self.undoStack:UndoAvailable()) then
		Controls.toolbar.tbUndo:Enable()
		mmf.Menu.Enable("EditUndo")
	else
		Controls.toolbar.tbUndo:Disable()
		mmf.Menu.Disable("EditUndo")
	end
	
	if (self.undoStack:RedoAvailable()) then
		Controls.toolbar.tbRedo:Enable()
		mmf.Menu.Enable("EditRedo")
	else
		Controls.toolbar.tbRedo:Disable()
		mmf.Menu.Disable("EditRedo")
	end

	-- Step individual tools
	self:StepSelect()
	self:StepDraw()
	self:StepErase()
	self:StepStamp()
	self:StepFill()
end

function Layer:ClearFloat ()
	local x1 = Controls.floatingBox.sx
	local y1 = Controls.floatingBox.sy
	local x2 = Controls.floatingBox.sx + Controls.floatingBox.dw - 1
	local y2 = Controls.floatingBox.sy + Controls.floatingBox.dh - 1
		
	local diffx = math.abs(Controls.floatingBox.px - Controls.floatingBox.sx)
	local diffy = math.abs(Controls.floatingBox.py - Controls.floatingBox.sy)
	
	self.map:ClearArea(x1 - diffx, y1 - diffy, x2 + diffx, y2 + diffy)
	self.map:PaintArea(x1 - diffx, y1 - diffy, x2 + diffx, y2 + diffy)
	
	return x1, y1
end

function Layer:PaintFloat ()
	local x1, y1 = self:ClearFloat()
	
	self.float:PaintAt(x1, y1)
end

function Layer:FloatUndo (commit)
	local x1 = Controls.floatingBox.sx
	local y1 = Controls.floatingBox.sy
	local x2 = Controls.floatingBox.sx + Controls.floatingBox.dw - 1
	local y2 = Controls.floatingBox.sy + Controls.floatingBox.dh - 1
	
	if (self.undoPool == nil) then
		self.undoPool = { data = {}, type = 'tile', sx = x1, sy = y1 }
	elseif (self.undoPool.sx == x1 and self.undoPool.sy == y1) then
		if (commit) then
			self.undoPool = nil
		end
		return
	end
		
	for i = x1, x2 do
		for j = y1, y2 do
			local d = i .. '_' .. j
			if (self.undoPool.data[d] == nil) then
				self.undoPool.data[d] = self.map:CopyTile(i, j) or {dx = i, dy = j}
			end
		end
	end
	
	if (commit) then
		self.undoStack:LogData(self.undoPool)
		self.undoPool = nil
		
		editor.SetModified()
	end
end

function Layer:StepSelect ()
	-- If out of mode, cleanup
	if (self.curTool ~= 'select') then
		if (Controls.floatingBox.type == 'select') then
			Controls.floatingBox:Release()
			if (Controls.floatingBox.active) then
				self:FloatUndo(1)				
				self:PaintFloat()
				self.map:MergeMap(self.float, Controls.floatingBox.sx, Controls.floatingBox.sy)
				self.float = {}
			end
			Controls.floatingBox:Stop()
			DisableCutCopy()
		end
		DisablePaste()
		return
	end
	
	-- Paste abilities
	if (self.clipboard) then
		EnablePaste()
	else
		DisablePaste()
	end

	if (Controls.clientArea.engaged and Controls.floatingBox.active == false) then
		Controls.floatingBox:Stop()
		DisableCutCopy()
		
		Controls.selectBox:StartSnapping(Cursor.canvas.xtile, Cursor.canvas.ytile)
	elseif (Controls.selectBox.active) then
		local coords = Controls.selectBox:Coords()
		if (coords.dw > 0 and coords.dh > 0) then
			Controls.floatingBox:StartSnapping(coords.sx, coords.sy, coords.dw, coords.dh)
			Controls.floatingBox.autoscroll = true
			Controls.floatingBox.type = 'select'
			
			EnableCutCopy()
			
			local x1 = Controls.floatingBox.sx
			local y1 = Controls.floatingBox.sy
			local x2 = Controls.floatingBox.sx + Controls.floatingBox.dw - 1
			local y2 = Controls.floatingBox.sy + Controls.floatingBox.dh - 1
			
			self:FloatUndo()
			
			-- Create float
			self.float = self.map:CopyArea(x1, y1, x2, y2)
			self.map:ClearArea(x1, y1, x2, y2)
			self.map:RemoveArea(x1, y1, x2, y2)
			self.float:PaintAt(x1, y1)
		end
		Controls.selectBox:Stop()
	end
	
	-- Cancel floating box
	if (Controls.clientArea.engagable and (Cursor.RDown or mmf.Keyboard.KeyDown(win.key.VK_RETURN)) and Controls.floatingBox.active) then
		self:FloatUndo(1)			
		self:PaintFloat()
		self.map:MergeMap(self.float, Controls.floatingBox.sx, Controls.floatingBox.sy)
		self.float = {}
		
		Controls.floatingBox:Stop()
		DisableCutCopy()
	end
	
	-- Delete floating box
	if (Controls.floatingBox.engaged == false and Controls.floatingBox.active and mmf.Keyboard.KeyDown(win.key.VK_DELETE)) then
		if (self.undoPool ~= nil) then
			self.undoStack:LogData(self.undoPool)
			self.undoPool = nil
		end
		self:ClearFloat()
		self.float = {}
		
		Controls.floatingBox:Stop()
		DisableCutCopy()
	end
	
	if (Controls.floatingBox.active == false) then
		return
	end
	
	Controls.floatingBox:TestEngage(Cursor.layer)
	
	-- Dragging
	if (Controls.clientArea.engagable and Controls.floatingBox.engagable) then
		Cursor:SetState('move')
		Controls.floatingBox:Release()
	elseif (Controls.clientArea.engaged and Controls.floatingBox.engaged) then
		Cursor:SetState('move')
		Controls.floatingBox:Grab()
		self:PaintFloat()
	elseif (Controls.floatingBox.engagable == false and Controls.floatingBox.engaged == false) then
		Controls.floatingBox:Release()
	end
end

function Layer:StepDraw ()
	-- If out of mode, exit
	if (self.curTool ~= 'draw') then
		return
	end
	
	-- Undo Logic
	if (Controls.clientArea.engaged == false and self.undoPool ~= nil) then
		self.undoStack:LogData(self.undoPool)
		self.undoPool = nil
	end
	
	-- If mouse is out of the client area, we have nothing to do
	if (Controls.clientArea.engagable == false and Controls.clientArea.engaged == false) then
		return
	end
	
	-- If cursor is outside of drawable region, we have nothing to do
	if (Cursor.canvas.xtile < 0 or Cursor.canvas.xtile >= Frame.frameWidth or
		Cursor.canvas.ytile < 0 or Cursor.canvas.ytile >= Frame.frameHeight) then
		return
	end
	
	-- If the mouse button isn't down, nothing's being done
	if (Controls.clientArea.engaged == false) then
		self.last_tx = -1
		self.last_ty = -1
		return
	end
	
	-- If we've already drawn this tile, don't repeat it
	if (Cursor.canvas.xtile == self.last_tx and Cursor.canvas.ytile == self.last_ty) then
		return
	end
	
	self.last_tx = Cursor.canvas.xtile
	self.last_ty = Cursor.canvas.ytile
	
	if (self.undoPool == nil) then
		self.undoPool = {
			data = {},
			type = 'tile',
		}
	end

	-- Add full block of tiles
	for i = 0, self.xSelectEnd - self.xSelectStart do
		for j = 0, self.ySelectEnd - self.ySelectStart do
			local dx = Cursor.canvas.xtile + i
			local dy = Cursor.canvas.ytile + j
			local sx = self.xSelectStart + i
			local sy = self.ySelectStart + j
			
			-- Log Undo Information
			local d = dx .. '_' .. dy
			if (self.undoPool.data[d] == nil) then
				self.undoPool.data[d] = self.map:CopyTile(dx, dy) or {dx = dx, dy = dy}
			end

			self.map:AddTile(dx, dy, sx, sy)
			self.map:PaintTile(dx, dy)
		end
	end
	
	editor.SetModified()
end

function Layer:StepErase ()
	-- If out of mode, exit
	if (self.curTool ~= 'erase') then
		return
	end
	
	-- Undo Logic
	if (Controls.clientArea.engaged == false and self.undoPool ~= nil) then
		self.undoStack:LogData(self.undoPool)
		self.undoPool = nil
	end
	
	-- If cursor is outside of drawable region, we have nothing to do
	if (Cursor.canvas.xtile < 0 or Cursor.canvas.xtile >= Frame.frameWidth or
		Cursor.canvas.ytile < 0 or Cursor.canvas.ytile >= Frame.frameHeight) then
		return
	end

	-- If the mouse button isn't down, nothing's being done
	if (Controls.clientArea.engaged == false) then
		self.last_tx = -1
		self.last_ty = -1
		return
	end
	
	-- If we've already erased this tile, don't repeat it
	if (Cursor.canvas.xtile == self.last_tx and Cursor.canvas.ytile == self.last_ty) then
		return
	end
	
	self.last_tx = Cursor.canvas.xtile
	self.last_ty = Cursor.canvas.ytile
	
	if (self.undoPool == nil) then
		self.undoPool = {
			data = {},
			type = 'tile',
		}
	end
	
	local dx = Cursor.canvas.xtile
	local dy = Cursor.canvas.ytile
	
	-- Log Undo Information
	local d = dx .. '_' .. dy
	if (self.undoPool.data[d] == nil) then
		self.undoPool.data[d] = self.map:CopyTile(dx, dy) or {dx = dx, dy = dy}
	end
	
	self.map:ClearTile(dx, dy)
	self.map:RemoveTile(dx, dy)
	
	editor.SetModified()
end

function Layer:StepStamp ()
	if (self.curTool ~= 'stamp') then
		if (Controls.floatingBox.type == 'stamp') then
			Controls.floatingBox:Release()
			Controls.floatingBox:Stop()
		end
		return
	end
	
	local x1, x2, y1, y2
	
	if (Controls.clientArea.engaged and Controls.floatingBox.active == false) then
		Controls.floatingBox:Stop()
		Controls.selectBox:StartSnapping(Cursor.canvas.xtile, Cursor.canvas.ytile)
	elseif (Controls.selectBox.active) then
		if (Controls.selectBox.dw > 0 and Controls.selectBox.dh > 0) then
			Controls.floatingBox:StartSnapping(Controls.selectBox.sx, Controls.selectBox.sy, Controls.selectBox.dw, Controls.selectBox.dh)
			Controls.floatingBox.type = 'stamp'
			
			x1 = Controls.floatingBox.sx
			y1 = Controls.floatingBox.sy
			x2 = Controls.floatingBox.sx + Controls.floatingBox.dw - 1
			y2 = Controls.floatingBox.sy + Controls.floatingBox.dh - 1

			self.float = self.map:CopyArea(x1, y1, x2, y2)
		end
		Controls.selectBox:Stop()
	end
	
	-- Cancel the stamp float when user clicks right mouse button in the client area
	if (Controls.clientArea.engagable and Cursor.RDown and Controls.floatingBox.active) then
		self:ClearFloat()
		Controls.floatingBox:Stop()
	end

	-- Paste the stamp float when the user clicks the left mouse button in the client area
	if (Controls.clientArea.engaged and Controls.floatingBox.active) then
		if (Cursor.canvas.xtile ~= self.last_tx or Cursor.canvas.ytile ~= self.last_ty) then
			self.undoPool = {
				data = {},
				type = 'tile',
			}
			
			for i = Controls.floatingBox.sx, Controls.floatingBox.sx + Controls.floatingBox.dw - 1 do
				for j = Controls.floatingBox.sy, Controls.floatingBox.sy + Controls.floatingBox.dh - 1 do
					local d = i .. '_' .. j
					local s = i - Controls.floatingBox.sx .. '_' .. j - Controls.floatingBox.sy
					if (self.float.map[s] ~= nil) then
						self.undoPool.data[d] = self.map:CopyTile(i, j) or {dx = i, dy = j}
					end
				end
			end
			
			self.map:MergeMap(self.float, Controls.floatingBox.sx, Controls.floatingBox.sy)

			self.undoStack:LogData(self.undoPool)
			self.undoPool = nil
			
			editor.SetModified()
		end
		
		self.last_tx = Cursor.canvas.xtile
		self.last_ty = Cursor.canvas.ytile
	else
		self.last_tx = -1
		self.last_ty = -1
	end

	-- If there's no float, there's nothing to do
	if (Controls.floatingBox.active == false) then
		return
	end
	
	-- If we do have a float and we're in the client area, use a stamp cursor
	if (Controls.clientArea.engagable or Controls.clientArea.engaged) then
		Cursor:SetState('stamp')
	end
	
	-- Keep the float following the cursor and draw it
	Controls.floatingBox:Grab(Controls.floatingBox.sx, Controls.floatingBox.sy)
	self:PaintFloat()
end

function Layer:StepFill ()
	if (self.curTool ~= 'fill') then
		return
	end
	
	if (Controls.clientArea.engaged == false) then
		self.latch = false
		return
	elseif (self.latch == true) then
		return
	end
	
	self.latch = true
	self.undoPool = {
		data = {},
		type = 'tile',
	}

	local target = self.map:UniqueID(Cursor.canvas.xtile, Cursor.canvas.ytile)
	local xrange = self.xSelectEnd - self.xSelectStart + 1
	local yrange = self.ySelectEnd - self.ySelectStart + 1
	
	-- Start flood fill algorithm
	local queue = Container.Queue:New()
	
	--local dx = cursor.canvas.xtile
	--local dy = cursor.canvas.ytile
	queue:Push({Cursor.canvas.xtile, Cursor.canvas.ytile})
	
	while (queue:Empty() == false) do
		local dx, dy = unpack(queue:Pop())
		local n = dx .. '_' .. dy
		
		--local merge = self.map:UniqueID(dx, dy)
		if (self.map:UniqueID(dx, dy) == target) then
			local edx = dx
			local wdx = dx
			
			while (wdx > 0) do
				if (self.map:UniqueID(wdx - 1, dy) ~= target) then
					break
				end
				wdx = wdx - 1
			end
			
			while (edx < Frame.frameWidth - 1) do
				if (self.map:UniqueID(edx + 1, dy) ~= target) then
					break
				end
				edx = edx + 1
			end
			
			for idx = wdx, edx do
				-- Color tile
				local d = idx .. '_' .. dy
				self.undoPool.data[d] = self.map:CopyTile(idx, dy) or {dx = idx, dy = dy}
				
				self.map:AddTile(idx, dy, self.xSelectStart + ((Cursor.canvas.xtile + idx) % xrange), self.ySelectStart + ((Cursor.canvas.ytile + dy) % yrange))
				self.map:PaintTile(idx, dy)
				
				local merge = self.map:UniqueID(idx, dy)
				local up = self.map:UniqueID(idx, dy - 1)
				local down = self.map:UniqueID(idx, dy + 1)
				
				if (up == target and up ~= merge and dy > 0) then
					queue:Push({idx, dy - 1})
				end
				if (down == target and down ~= merge and dy < Frame.frameHeight - 1) then
					queue:Push({idx, dy + 1})
				end
			end
		end
	end
	
	self.undoStack:LogData(self.undoPool)
	self.undoPool = nil
	
	editor.SetModified()
end

function Layer:UnRedo (stack1)
	local stack2 = {
		data = {},
		type = stack1.type,
	}

	for k,v in pairs(stack1.data) do
		stack2.data[k] = self.map:CopyTile(v.dx, v.dy) or {dx = v.dx, dy = v.dy}
		self.map:RemoveTile(v.dx, v.dy)
		if (v.cnt) then
			self.map:MergeTile(v, v.dx, v.dy)
		end
		self.map:ClearTile(v.dx, v.dy)
		self.map:PaintTile(v.dx, v.dy)
	end
	
	return stack2
end

function Layer:Undo ()
	if (self.undoStack:UndoAvailable() == false) then
		return
	end
	
	local redo = self:UnRedo(self.undoStack:Undo())
	self.undoStack:LogRedo(redo)
	
	editor.SetModified()
end

function Layer:Redo ()
	if (self.undoStack:RedoAvailable() == false) then
		return
	end

	local undo = self:UnRedo(self.undoStack:Redo())
	self.undoStack:LogUndo(undo)
	
	editor.SetModified()
end

-- ----------------------------------------------------------------------------
-- Tile Manager Overrides

-- The manager doesn't know the difference between an editor and a runtime, so
-- we override these two system-level functions to actually do dirty work
-- inside MMF.

function Tile._PaintTile (dx, dy, sx, sy)
	local dx = Controls.clientArea.reserved + dx * Current.tileWidth
	local dy = Controls.clientArea.reserved + dy * Current.tileHeight
	local sw = Current.tileWidth
	local sh = Current.tileHeight
	local sx = sx * (Current.tileWidth + Current.xSpacing) + Current.xOrigin
	local sy = sy * (Current.tileHeight + Current.ySpacing) + Current.yOrigin
	
	Current.controls.tilePaletteSource:PasteBG(dx, dy, sx, sy, sw, sh)
end

function Tile._ClearTile (dx, dy)
	local dx = Controls.clientArea.reserved + dx * Current.tileWidth
	local dy = Controls.clientArea.reserved + dy * Current.tileHeight

	mmf.Frame.DeleteBG(Conf.baseLayer + Current.order, dx, dy)
end