local util		= require "./lua/util"
local tile		= require "./lua/tile"

module (..., package.seeall)

Import = { predata = { }, postdata = { } }
Export = { }

Info = { }

-- ----------------------------------------------------------------------------

function DoImport (filedata)
	return Import:Header(filedata)
end

function DoExport ()
	return Export:Header()
end

-- ----------------------------------------------------------------------------

function Import:PostImport ()
	for k,v in pairs(self.postdata) do
		for k2,v2 in pairs(v) do
			editor.layers[k][k2] = v2
		end
	end
	
	editor.vars.layer:Reload()
	EditorUpdate()
end

-- ----------------------------------------------------------------------------

Info.Header = {
	ident = "TLEV",
	ver = 0,
	szBlock = 26,
	szIndex = 8,
	
	compatIdent = "TLV2",
	numBlocks = 3,
}

function Import:Header (filedata)
	local file_len = string.len(filedata)
	if (file_len < Info.Header.szBlock) then
		return false
	end

	local descriptor = string.sub(filedata, 1, 8)
	
	local mime = string.sub(descriptor, 1, 4)
	if (mime ~= Info.Header.ident) then
		return false
	end
	
	local ver = util.unu16(string.sub(descriptor, 5, 6))
	local hsz = util.unu16(string.sub(descriptor, 7, 8))
	
	local headerBlock = string.sub(filedata, 1, hsz)
	
	local fileSize = util.unu32(string.sub(headerBlock, 9, 12))
	local offsetIndexBlock = util.unu32(string.sub(headerBlock, 13, 16)) + 1
	local numIndexBlock = util.unu32(string.sub(headerBlock, 17, 20))
	local szIndexBlock = util.unu16(string.sub(headerBlock, 21, 22))

	local compatID = string.sub(headerBlock, 23, 26)
	if (compatID ~= Info.Header.compatIdent) then
		return false
	end
	
	EditorReset()
	self.predata = { }
	self.postdata = { }

	local indexBlockSet = string.sub(filedata, offsetIndexBlock, offsetIndexBlock + numIndexBlock * szIndexBlock)
	for i = 1, numIndexBlock do
		local indexBase = 1 + (i - 1) * szIndexBlock
		local indexBlock = string.sub(indexBlockSet, indexBase, indexBase + szIndexBlock)
		
		local offsetBlock = util.unu32(string.sub(indexBlock, 1, 4)) + 1
		local szBlock = util.unu32(string.sub(indexBlock, 5, 8))
		local block = string.sub(filedata, offsetBlock, offsetBlock + szBlock)
		
		local ident = string.sub(block, 1, 4)
		if (ident == "FRAM") then
			self:Frame(block)
		elseif (ident == "LAYE") then
			self:Layers(block)
		elseif (ident == "TILE") then
			self:TileLayers(block)
		end
	end
	
	editor.InitLayers(self.predata)
	EditorReInit()
	
	self:PostImport()
	
	return true
end

function Export:Header ()
	local data = ""
	local indexData = ""
	
	local blockFrame = self:Frame()
	local blockLayers = self:Layers()
	local blockTileLayers = self:TileLayers()
	
	local offsetFrame = Info.Header.szBlock + (Info.Header.numBlocks * Info.Header.szIndex)
	local offsetLayers = offsetFrame + string.len(blockFrame)
	local offsetTileLayers = offsetLayers + string.len(blockLayers)
	
	local filesize = offsetTileLayers + string.len(blockTileLayers)
	
	indexData = indexData .. util.u32(offsetFrame) .. util.u32(string.len(blockFrame))
	indexData = indexData .. util.u32(offsetLayers) .. util.u32(string.len(blockLayers))
	indexData = indexData .. util.u32(offsetTileLayers) .. util.u32(string.len(blockTileLayers))
	
	data = Info.Header.ident .. util.u16(Info.Header.ver) .. util.u16(Info.Header.szBlock) .. util.u32(filesize)
		.. util.u32(Info.Header.szBlock) .. util.u32(Info.Header.numBlocks) .. util.u16(Info.Header.szIndex) .. Info.Header.compatIdent
		
	return data .. indexData .. blockFrame .. blockLayers .. blockTileLayers
end

-- ----------------------------------------------------------------------------

Info.Frame = {
	ident = "FRAM",
	ver = 0,
	szBlock = 52,
}

function Import:Frame (blockData)
	local mime = string.sub(blockData, 1, 4)
	local ver = util.unu16(string.sub(blockData, 5, 6))
	local szBlock = util.unu16(string.sub(blockData, 7, 8))
	
	local header = string.sub(blockData, 1, szBlock)
	
	local offsetStr = util.unu32(string.sub(header, 9, 12)) + 1
	local numStr = util.unu32(string.sub(header, 13, 16))
	
	local strBlockSet = string.sub(blockData, offsetStr, string.len(blockData))
	
	local strTable = { }
	local strIndex = 1
	for i = 1, numStr do
		local strLen = util.unu32(string.sub(strBlockSet, strIndex, strIndex + 3))
		strTable[i] = string.sub(strBlockSet, strIndex + 4, strIndex + 3 + strLen)
		strIndex = strIndex + 4 + strLen
	end
	
	Frame.frameWidth = util.unu32(string.sub(header, 17, 20))
	Frame.frameHeight = util.unu32(string.sub(header, 21, 24))
	Frame.backgroundColor = util.unu32(string.sub(header, 25, 28))
	
	Frame.levelName = strTable[util.unu32(string.sub(header, 29, 32))]
	Frame.author = strTable[util.unu32(string.sub(header, 33, 36))]
	Frame.key = strTable[util.unu32(string.sub(header, 37, 40))]
	
	Frame.dateCreated = util.unu32(string.sub(header, 41, 44))
	Frame.dateModified = util.unu32(string.sub(header, 45, 48))
	Frame.tileWidth = util.unu16(string.sub(header, 49, 50))
	Frame.tileHeight = util.unu16(string.sub(header, 51, 52))
end

function Export:Frame ()
	local data = ""
	
	local strTable = { }
	local strData = ""
	
	strTable[1] = util.u32(string.len(Frame.levelName)) .. Frame.levelName
	strTable[2] = util.u32(string.len(Frame.author)) .. Frame.author
	strTable[3] = util.u32(string.len(Frame.key)) .. Frame.key
	
	for i,v in ipairs(strTable) do
		strData = strData .. v
	end
	
	local created = Frame.dateCreated
	if (created == 0 or created == nil) then
		created = os.time()
	end
	
	local modified = os.time()
	
	local offsetStr = Info.Frame.szBlock
	
	data = Info.Frame.ident .. util.u16(Info.Frame.ver) .. util.u16(Info.Frame.szBlock) .. util.u32(offsetStr)
		.. util.u32(#strTable) .. util.u32(Frame.frameWidth) .. util.u32(Frame.frameHeight) .. util.u32(Frame.backgroundColor)
		.. util.u32(1) .. util.u32(2) .. util.u32(3) .. util.u32(created) .. util.u32(modified) .. util.u16(Frame.tileWidth)
		.. util.u16(Frame.tileHeight)
		
	return data .. strData
end

-- ----------------------------------------------------------------------------

Info.Layers = {
	ident = "LAYE",
	ver = 0,
	szBlock = 26,
	szLayerOrder = 6,
}

function Import:Layers (blockData) 
	local mime = string.sub(blockData, 1, 4)
	local ver = util.unu16(string.sub(blockData, 5, 6))
	local szBlock = util.unu16(string.sub(blockData, 7, 8))
	
	local header = string.sub(blockData, 1, szBlock)
	
	local offsetOrder = util.unu32(string.sub(header, 9, 12)) + 1
	local numOrder = util.unu32(string.sub(header, 13, 16))
	local szOrder = util.unu16(string.sub(header, 17, 18))
	local offsetStr = util.unu32(string.sub(header, 19, 22)) + 1
	local numStr = util.unu32(string.sub(header, 23, 26))
	
	local orderBlockSet = string.sub(blockData, offsetOrder, offsetOrder + numOrder * szOrder)
	local strBlockSet = string.sub(blockData, offsetStr, string.len(blockData))
	
	local strTable = { }
	local strIndex = 1
	for i = 1, numStr do
		local strLen = util.unu32(string.sub(strBlockSet, strIndex, strIndex + 3))
		strTable[i] = string.sub(strBlockSet, strIndex + 4, strIndex + 3 + strLen)
		strIndex = strIndex + 4 + strLen
	end
	
	local typeMap = { 'tile', 'object', 'backdrop' }
	
	-- Still need control INIT .. perhaps not here

	for i = 1, numOrder do
		local baseIndex = (i - 1) * szOrder + 1
		local orderBlock = string.sub(orderBlockSet, baseIndex, baseIndex + szOrder)
		
		local type = util.unu16(string.sub(orderBlock, 1, 2))
		local index = util.unu16(string.sub(orderBlock, 3, 4))
		local strIndex = util.unu16(string.sub(orderBlock, 5, 6))
		
		self.predata[i] = { }
		self.predata[i].order = i
		self.predata[i].instOrder = index
		self.predata[i].type = typeMap[type]
		self.predata[i].name = strTable[strIndex]
	end
end

function Export:Layers ()
	local data = ""
	local orderData = ""
	
	local strTable = { }
	local strData = ""
	
	local instcount = { }
	
	for i,layer in ipairs(editor.layers) do
		if (instcount[layer.layerType] == nil) then
			instcount[layer.layerType] = 1
		else
			instcount[layer.layerType] = instcount[layer.layerType] + 1
		end
		
		strTable[#strTable+1] = util.u32(string.len(layer.name)) .. layer.name
		
		orderData = orderData .. util.u16(layer.layerType) .. util.u16(instcount[layer.layerType]) .. util.u16(#strTable)
	end
	
	for i,v in ipairs(strTable) do
		strData = strData .. v
	end
	
	local offsetOrder = Info.Layers.szBlock
	local offsetStr = offsetOrder + (#editor.layers * Info.Layers.szLayerOrder)
	
	data = Info.Layers.ident .. util.u16(Info.Layers.ver) .. util.u16(Info.Layers.szBlock) .. util.u32(offsetOrder)
		.. util.u32(#editor.layers) .. util.u16(Info.Layers.szLayerOrder) .. util.u32(offsetStr) .. util.u32(#strTable)
		
	return data .. orderData .. strData
end

-- ----------------------------------------------------------------------------

Info.TileLayers = {
	ident = "TILE",
	ver = 0,
	szBlock = 46,
	szTileLayer = 24,
	szTileInst = 12,
	szStackEntry = 4,
	
	type = 1,
}

function Import:TileLayers (blockData)
	local mime = string.sub(blockData, 1, 4)
	local ver = util.unu16(string.sub(blockData, 5, 6))
	local szBlock = util.unu16(string.sub(blockData, 7, 8))
	
	local header = string.sub(blockData, 1, szBlock)
	
	local offsetLayer = util.unu32(string.sub(header, 9, 12)) + 1
	local offsetInst = util.unu32(string.sub(header, 13, 16)) + 1
	local offsetStack = util.unu32(string.sub(header, 17, 20)) + 1
	local numLayer = util.unu32(string.sub(header, 21, 24))
	local numInst = util.unu32(string.sub(header, 25, 28))
	local numStack = util.unu32(string.sub(header, 29, 32))
	local szLayer = util.unu16(string.sub(header, 33, 34))
	local szInst = util.unu16(string.sub(header, 35, 36))
	local szStack = util.unu16(string.sub(header, 37, 38))
	local offsetStr = util.unu32(string.sub(header, 39, 42)) + 1
	local numStr = util.unu32(string.sub(header, 43, 46))
	
	local layerBlockSet = string.sub(blockData, offsetLayer, offsetLayer + numLayer * szLayer)
	local instBlockSet = string.sub(blockData, offsetInst, offsetInst + numInst * szInst)
	local stackBlockSet = string.sub(blockData, offsetStack, offsetStack + numStack * szStack)
	local strBlockSet = string.sub(blockData, offsetStr, string.len(blockData))
	
	local strTable = { }
	local strIndex = 1
	for i = 1, numStr do
		local strLen = util.unu32(string.sub(strBlockSet, strIndex, strIndex + 3))
		strTable[i] = string.sub(strBlockSet, strIndex + 4, strIndex + 3 + strLen)
		strIndex = strIndex + 4 + strLen
	end
	
	local typeMap = { 'normal', 'collision', 'platform' }
	
	-- Stack entries
	local stackTable = { }
	for i = 1, numStack do
		local baseIndex = (i - 1) * szStack + 1
		local stackBlock = string.sub(stackBlockSet, baseIndex, baseIndex + szStack)
		
		stackTable[i] = {
			sx = util.unu16(string.sub(stackBlock, 1, 2)),
			sy = util.unu16(string.sub(stackBlock, 3, 4))
		}
	end
	
	-- Tile entries
	local instTable = { }
	for i = 1, numInst do
		local baseIndex = (i - 1) * szInst + 1
		local instBlock = string.sub(instBlockSet, baseIndex, baseIndex + szInst)
		
		instTable[i] = {
			layerIndex = util.unu16(string.sub(instBlock, 1, 2)),
			stackIndex = util.unu32(string.sub(instBlock, 3, 6)),
			numStack = util.unu16(string.sub(instBlock, 7, 8)),
			dx = util.unu16(string.sub(instBlock, 9, 10)),
			dy = util.unu16(string.sub(instBlock, 11, 12))
		}
	end
	
	-- Layer entries
	local layerTable = { }
	for i = 1, numLayer do
		local baseIndex = (i - 1) * szLayer + 1
		local layerBlock = string.sub(layerBlockSet, baseIndex, baseIndex + szLayer)
		
		layerTable[i] = {
			type = util.unu16(string.sub(layerBlock, 1, 2)),
			numInst = util.unu32(string.sub(layerBlock, 3, 6)),
			numStack = util.unu32(string.sub(layerBlock, 7, 10)),
			tileWidth = util.unu16(string.sub(layerBlock, 11, 12)),
			tileHeight = util.unu16(string.sub(layerBlock, 13, 14)),
			xSpacing = util.unu16(string.sub(layerBlock, 15, 16)),
			ySpacing = util.unu16(string.sub(layerBlock, 17, 18)),
			xOffset = util.unu16(string.sub(layerBlock, 19, 20)),
			yOffset = util.unu16(string.sub(layerBlock, 21, 22)),
			tileset = strTable[util.unu16(string.sub(layerBlock, 23, 24))]
		}
	end
	
	-- Put layers back together
	local tilemaps = { }
	for i = 1, numLayer do
		local xMax = math.ceil(Frame.frameWidth / layerTable[i].tileWidth)
		local yMax = math.ceil(Frame.frameHeight / layerTable[i].tileHeight)
		tilemaps[i] = tile.TileMap:new({xMax = xMax, yMax = yMax})
	end
	
	for i = 1, numInst do
		for j = instTable[i].stackIndex + 1, instTable[i].stackIndex + instTable[i].numStack do
			tilemaps[instTable[i].layerIndex]:AddTile(instTable[i].dx, instTable[i].dy, stackTable[j].sx, stackTable[j].sy)
		end
	end

	for i, layer in ipairs(self.predata) do
		if (layer.type == 'tile') then
			self.predata[i].subtype = typeMap[layerTable[layer.instOrder].type]
			self.postdata[i] = layerTable[layer.instOrder]
			self.postdata[i].map = tilemaps[layer.instOrder]
		end
	end
end

function Export:TileLayers ()
	local data = ""
	local layerData = ""
	local instData = ""
	local stackData = ""
	
	local layerCount = 0
	local instCount = 0
	local stackCount = 0
	
	local strTable = { }
	local strData = ""
	
	local tileLayers = { }
	
	for i,layer in ipairs(editor.layers) do
		if (layer.layerType == Info.TileLayers.type) then
			tileLayers[#tileLayers+1] = layer
		end
	end
	
	for i,layer in ipairs(tileLayers) do
		local innerTileCount = 0
		local innerStackCount = 0
		
		for k,tile in pairs(layer.map.map) do
			local tileStackCount = 0
			local tileStackIndex = stackCount
			
			local stack = util.paramsort(tile.tile, 'depth')
			for j,se in ipairs(stack) do
				stackData = stackData .. util.u16(se.sx) .. util.u16(se.sy)
				
				tileStackCount = tileStackCount + 1
				innerStackCount = innerStackCount + 1
				stackCount = stackCount + 1
			end
			
			instData = instData .. util.u16(i) .. util.u32(tileStackIndex) .. util.u16(tileStackCount)
					.. util.u16(tile.dx) .. util.u16(tile.dy)
			
			innerTileCount = innerTileCount + 1
			instCount = instCount + 1
		end
		
		strTable[#strTable+1] = util.u32(string.len(layer.tileset)) .. layer.tileset

		layerData = layerData .. util.u16(layer.typeid) .. util.u32(innerTileCount) .. util.u32(innerStackCount)
				 .. util.u16(layer.tileWidth) .. util.u16(layer.tileHeight) .. util.u16(layer.xSpacing) .. util.u16(layer.ySpacing)
				 .. util.u16(layer.xOrigin) .. util.u16(layer.yOrigin) .. util.u16(#strTable)
				 
		layerCount = layerCount + 1
	end
	
	for i,v in ipairs(strTable) do
		strData = strData .. v
	end
	
	local offsetTileLayer = Info.TileLayers.szBlock
	local offsetTileInst = offsetTileLayer + (layerCount * Info.TileLayers.szTileLayer)
	local offsetTileStack = offsetTileInst + (instCount * Info.TileLayers.szTileInst)
	local offsetStr = offsetTileStack + (stackCount * Info.TileLayers.szStackEntry)
	
	data = Info.TileLayers.ident .. util.u16(Info.TileLayers.ver) .. util.u16(Info.TileLayers.szBlock) .. util.u32(offsetTileLayer)
		.. util.u32(offsetTileInst) .. util.u32(offsetTileStack) .. util.u32(layerCount) .. util.u32(instCount)
		.. util.u32(stackCount) .. util.u16(Info.TileLayers.szTileLayer) .. util.u16(Info.TileLayers.szTileInst)
		.. util.u16(Info.TileLayers.szStackEntry) .. util.u32(offsetStr) .. util.u32(#strTable)
	
	return data .. layerData .. instData .. stackData .. strData
end