-- Tile Management
-- ----------------------------------------------------------------------------

local util = require "./lua/util"

local pairs = pairs
local ipairs = ipairs
local setmetatable = setmetatable

module (...)

-- ----------------------------------------------------------------------------

TileMap = {
	xMax = 0,
	yMax = 0,
	map = { },
}

function TileMap:new (obj)
	obj = obj or {}
	if (obj.map == nil) then obj.map = {} end
	
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function TileMap:UniqueID (dx, dy)
	local d = dx .. '_' .. dy
	if (self.map[d] == nil) then
		return nil
	end
	
	return self.map[d].uniqueid
end

function TileMap:AddTile (dx, dy, sx, sy)
	-- Check Destination bounds
	if (dx < 0 or dy < 0 or dx >= self.xMax or dy >= self.yMax) then
		return
	end
	
	-- Check Source bounds
	if (sx < 0 or sy < 0) then
		return
	end
	
	-- Make IDs
	local d = dx .. '_' .. dy

	-- Create an entry at this destination if no tiles exist there yet
	if (self.map[d] == nil) then
		self.map[d] = TileMapEntry:new( {parent = self, dx = dx, dy = dy} )
	end
	
	self.map[d]:Add(sx, sy)
end

function TileMap:MergeTile (tile, dx, dy)
	-- Check Destination bounds
	if (dx < 0 or dy < 0 or dx >= self.xMax or dy >= self.yMax) then
		return
	end
	
	local d = dx .. '_' .. dy
	if (self.map[d] == nil) then
		self.map[d] = TileMapEntry:new( {parent = self, dx = dx, dy = dy} )
	end
	
	self.map[d]:Merge(tile)
end

function TileMap:MergeMap(map, dx, dy)
	if (dx == nil) then
		dx = 0
	end
	if (dy == nil) then
		dy = 0
	end
	
	for k,v in pairs(map.map) do
		self:MergeTile(v, dx + v.dx, dy + v.dy)
	end
end

function TileMap:RemoveTile (dx, dy)
	local d = dx .. '_' .. dy
	if (self.map[d] == nil) then
		return
	end

	self.map[d] = nil
end

function TileMap:RemoveArea (dx1, dy1, dx2, dy2)
	for i = dx1, dx2 do
		for j = dy1, dy2 do
			self:RemoveTile(i, j)
		end
	end
end

function TileMap:CopyTile (dx, dy)
	local d = dx .. '_' .. dy
	if (self.map[d] == nil) then
		return nil
	end

	return TileMapEntry:new( {
		parent = self.map[d].parent,
		dx = self.map[d].dx,
		dy = self.map[d].dy,
		cnt = self.map[d].cnt,
		depth = self.map[d].depth,
		tile = util.clone(self.map[d].tile),
	} )
end

function TileMap:CopyArea (dx1, dy1, dx2, dy2)
	local xMax = dx2 - dx1
	local yMax = dy2 - dy1
	local area = TileMap:new( {xMax = xMax, yMax = yMax} )
	
	for i = 0, xMax do
		for j = 0, yMax do
			local d = i .. '_' .. j
			area.map[d] = self:CopyTile(dx1 + i, dy1 + j)
			if (area.map[d] ~= nil) then
				area.map[d].dx = i
				area.map[d].dy = j
			end
		end
	end
	
	return area
end

function TileMap:ClearTile (dx, dy)
	_ClearTile(dx, dy)
end

function TileMap:ClearArea (dx1, dy1, dx2, dy2)
	for i = dx1, dx2 do
		for j = dy1, dy2 do
			self:ClearTile(i, j)
		end
	end
end

function TileMap:ClearMap ()
	for k, v in pairs(self.map) do
		v:Clear()
	end
end

function TileMap:PaintTile (dx, dy)
	local d = dx .. '_' .. dy
	if (self.map[d] == nil) then
		return
	end
	
	self.map[d]:Paint()
end

function TileMap:PaintArea (dx1, dy1, dx2, dy2)
	for i = dx1, dx2 do
		for j = dy1, dy2 do
			self:PaintTile(i, j)
		end
	end
end

function TileMap:PaintAt (dx, dy)
	for k,v in pairs(self.map) do
		v:PaintAt(v.dx + dx, v.dy + dy)
	end
end

-- ----------------------------------------------------------------------------

TileMapEntry = {
	parent = nil,
	dx = 0,
	dy = 0,
	cnt = 0,
	depth = 0,
	uniqueid = nil,
	tile = { },
}

function TileMapEntry:new (obj)
	obj = obj or {}
	if (obj.tile == nil) then obj.tile = {} end
	
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function TileMapEntry:UniqueID ()
	local stack = util.paramsort(self.tile, 'depth')
	local uid = ''
	
	for i,v in ipairs(stack) do
		uid = uid .. v.sx .. '_' .. v.sy .. '_'
	end
	
	if (uid == '') then
		return nil
	end
	
	return uid
end

function TileMapEntry:MaxDepth ()
	local md = 0
	for k,v in pairs(self.tile) do
		if (v.depth > md) then
			md = v.depth
		end
	end
	return md
end

function TileMapEntry:Add (sx, sy)
	local s = sx .. '_' .. sy
	
	-- If this source tile doesn't exist on the stack, add it.  If it exists, move
	-- it to the top.  If it's already on the top, ignore it and don't redraw.
	if (self.tile[s] == nil) then
		self.cnt = self.cnt + 1
		self.depth = self.depth + 1
		self.tile[s] = TileStackEntry:new( {sx = sx, sy = sy, depth = self.depth} )
	elseif (self.tile[s].depth < self.depth) then
		self.depth = self.depth + 1
		self.tile[s].depth = self.depth
	end
	
	self.uniqueid = self:UniqueID()
end

function TileMapEntry:Merge (tile)
	local stack = util.paramsort(tile.tile, 'depth')
	for i,v in ipairs(stack) do
		self:Add(v.sx, v.sy)
	end
end

function TileMapEntry:Remove (sx, sy)
	local s = sx .. '_' .. sy
	if (self.tile[s] ~= nil) then
		self.tile[s] = nil
		self.cnt = self.cnt - 1
		self.depth = self:MaxDepth()
	end
	
	self.uniqueid = self:UniqueID()
end

function TileMapEntry:RemoveTop ()
	local sx, sy = nil
	for k,v in pairs(self.tile) do
		if (v.depth == self.depth) then
			sx = v.sx
			sy = v.sy
		end
	end
	
	self:Remove(sx, sy)
end

function TileMapEntry:RemoveAll ()
	self.cnt = 0
	self.depth = 0
	self.tile = { }
	self.uniqueid = nil
end

function TileMapEntry:Clear ()
	_ClearTile(self.dx, self.dy)
end

function TileMapEntry:Paint ()
	self:PaintAt(self.dx, self.dy)
end

function TileMapEntry:PaintAt (dx, dy)
	local stack = util.paramsort(self.tile, 'depth')
	for i,v in ipairs(stack) do
		_PaintTile(dx, dy, v.sx, v.sy)
	end
end

-- ----------------------------------------------------------------------------

TileStackEntry = {
	sx = 0,
	sy = 0,
	depth = 0,
}

function TileStackEntry:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	return obj
end

-- ----------------------------------------------------------------------------

function _PaintTile (dx, dy, sx, sy) end

function _ClearTile (dx, dy) end